import math
import shutil
from datetime import datetime
from time import time
# from pytorch_model_summary import summary

import matplotlib.pyplot as plt
# from matplotlib.ticker import FormatStrFormatter
import numpy as np
import torch.utils.data
import argparse
import os
import json
import torch.nn.functional as F
import torchvision.transforms as T

# from tqdm import tqdm
plt.rcParams["figure.figsize"] = (18, 5)
# plt.rcParams["figure.figsize"] = (24,10)

from modules.file_utils import FileUtils
from modules.csv_utils_2 import CsvUtils2
from modules.celeb_dataset import CelebDataset
from modules.metric_functions import IoU, dice_coeficient, make_figures

#
parser = argparse.ArgumentParser(add_help=False)
print(parser)


parser.add_argument('-datasource_directory', default='./celeb_hq_dataset', type=str)
parser.add_argument('-mmap_filename', default='data.mmap', type=str)
parser.add_argument('-json_filename', default='data.json', type=str)
parser.add_argument('-learning_rate', default=1e-3, type=float)
parser.add_argument('-batch_size', default=16, type=int)
parser.add_argument('-epochs', default=100, type=int)
parser.add_argument('-num_layers', default=3, type=int)
parser.add_argument('-coef_num_features', default=3, type=int)
parser.add_argument('-sequence_name', default='seq_rand_rot', type=str)
parser.add_argument('-run_name', default='run', type=str)
parser.add_argument('-is_cuda', default=True, type=lambda x: (str(x).lower() == 'true'))
# parser.add_argument('-model', default='2_model_pre_trained_fcn', type=str)
parser.add_argument('-model', default='1_model_unet', type=str)
parser.add_argument('-save_figs', default=True, type=lambda x: (str(x).lower() == 'true'))
parser.add_argument('-early_stopping_patience', default=10, type=int)
parser.add_argument('-loss_tolerance', default=1e-5, type=float)
# parser.add_argument('-cuda_parallel', default=True, type=lambda x: (str(x).lower() == 'true'))

args, other_args = parser.parse_known_args()
print(args)

if not torch.cuda.is_available() or not args.is_cuda:
    args.device = 'cpu'
    args.is_cuda = False
    print('cuda not available')
else:
    args.device = 'cuda'
    print(f'cuda devices: {torch.cuda.device_count()}')

# Begin code
path_sequence = f'./results/{args.sequence_name}'
# args.run_name += ('-' + datetime.utcnow().strftime(f'%y-%m-%d--%H-%M-%S-%f'))
path_run = f'./results/{args.sequence_name}/{args.run_name}'
FileUtils.createDir(path_run)

# path_artificats = f'./artifacts/{args.sequence_name}/{args.run_name}'
# FileUtils.createDir(path_artificats)
# print(f'created {path_artificats}')
# FileUtils.writeJSON(f'{path_run}/args.json', args.__dict__)
# print('file utils done')

CsvUtils2.create_global(path_sequence)
CsvUtils2.create_local(path_sequence, args.run_name)

print('csv utils setup done')

dataset_full = CelebDataset(args)
len_dataset = len(dataset_full)
print(len_dataset)
len_train = int(len_dataset * 0.8)
len_test = len_dataset - len_train  # int(len_dataset*0.2)
print(len_train, len_test)

torch.random.manual_seed(0)
dataset_train, dataset_test = torch.utils.data.random_split(dataset_full, [len_train, len_test])  # propotions 80/20
torch.random.manual_seed(int(time()))

# data loaders
data_loader_train = torch.utils.data.DataLoader(
    dataset=dataset_train,
    batch_size=args.batch_size,
    shuffle=True,
    drop_last=(len(dataset_train) % args.batch_size == 1)
)
data_loader_test = torch.utils.data.DataLoader(
    dataset=dataset_test,
    batch_size=args.batch_size,
    shuffle=False,
    drop_last=(len(dataset_test) % args.batch_size == 1)
)
last_batch_size = len(dataset_test) % args.batch_size if len(dataset_test) % args.batch_size > 1 else args.batch_size
print('last_batch_size', last_batch_size)

# print(data_loader_train[0])

input_channel_names = dataset_full.getInputChannels()
class_names = dataset_full.getClassList()
print(f'input_channel_names = {input_channel_names}')
print(f'class_names = {class_names}')
# data_channels_dict = {'input': input_channel_names,
#                       'output': class_names}

classes = len(class_names)
args.classes = classes
Model = getattr(__import__('models.' + args.model, fromlist=['Model']), 'Model')
model = Model(classes=classes, args=args).to(args.device)
is_data_parallel = False
if torch.cuda.device_count() > 1:
    is_data_parallel = True
    model = torch.nn.DataParallel(model, dim=0)

optimizer = torch.optim.Adam(model.parameters(), lr=args.learning_rate)

params = list(model.parameters(recurse=True))
param_total = np.sum([np.prod(it.shape) for it in params])
print(f'param_total: {param_total}')
args.model_params = int(param_total)


args.input_channels = input_channel_names
args.output_channels = class_names
FileUtils.writeJSON(f'{path_run}/args.json', args.__dict__)

metrics = {}
metrics_mean_dict = {}
for stage in ['train', 'test']:
    for metric in [
        'loss',
        # 'dice_coef',
        'iou_coef'
    ]:
        metrics[f'{stage}_{metric}'] = []
        metrics_mean_dict[f'{stage}_{metric}'] = 1e20
        if (stage == 'test') and (metric == 'iou_coef'):
            for idx_class, class_name in enumerate(class_names):
                metrics[f'{stage}_{metric}_{class_name}'] = []
                metrics_mean_dict[f'{stage}_{metric}_{class_name}'] = 1e20

# print(metrics)
# print(metrics_mean_dict)

best_test_loss = 1e20
best_test_iou = 0
best_test_counter = 0
best_test_iou_list = np.zeros(classes)


for epoch in range(1, args.epochs):
    for data_loader in [data_loader_train, data_loader_test]:
        metrics_epoch = {key: [] for key in metrics.keys()}
        # for stage in ['train', 'test']:
        #     for idx_class, class_name in enumerate(class_names):
        #         metrics_epoch[f'{stage}_iou_coef_{class_name}'] = []

        stage = 'train'
        if data_loader == data_loader_test:
            stage = 'test'
        # for x, y in tqdm(data_loader):
        for x, y in data_loader:
            # print(x.shape, y.shape)
            x = x.to(args.device)
            y = y.to(args.device)

            if stage == 'test':
                model = model.eval()
                with torch.no_grad():
                    y_prim = model.forward(x)
                    # print(y_prim.shape)
                    dice_coef = dice_coeficient(y_prim, y)
                    loss = 1 - dice_coef
                    iou_coef, iou_classes = IoU(y_prim, y)
                    # if loss < best_test_loss:
                    #     best_test_loss = loss.item()
                    #     best_test_iou = iou_coef.item()
            else:
                model = model.train()
                y_prim = model.forward(x)
                # print(y_prim.shape)
                dice_coef = dice_coeficient(y_prim, y)
                loss = 1 - dice_coef
                iou_coef, iou_classes = IoU(y_prim, y)

                # for idx_class, class_name in enumerate(class_names):
                #     print(f'{idx_class} {iou_classes[idx_class].item()}')

            if data_loader == data_loader_train:
                loss.backward()
                optimizer.step()
                optimizer.zero_grad()

            loss = loss.cpu()
            y_prim = y_prim.cpu()
            x = x.cpu()
            y = y.cpu()

            np_x = x.data.numpy()
            np_y_prim = y_prim.data.numpy()
            np_y = y.data.numpy()

            metrics_epoch[f'{stage}_loss'].append(loss.item())
            # metrics_epoch[f'{stage}_dice_coef'].append(dice_coef.item())
            metrics_epoch[f'{stage}_iou_coef'].append(iou_coef.item())
            if stage == 'test':
                for idx_class, class_name in enumerate(class_names):
                    metrics_epoch[f'{stage}_iou_coef_{class_name}'].append(iou_classes[idx_class].item())


        metrics_strs = []
        for key in metrics_epoch.keys():
            if stage in key:
                value = np.mean(metrics_epoch[key])
                metrics[key].append(value)
                metrics_strs.append(f'{key}: {round(value, 2)}')
                metrics_mean_dict[key] = value

                # if key == 'test_loss':
                #     if best_test_loss > value:
                #         best_test_loss = value
                #         best_test_iou = metrics_mean_dict['test_iou_coef']

    best_test_counter += 1
    if metrics_mean_dict['test_loss'] < best_test_loss:
        if metrics_mean_dict['test_loss'] < best_test_loss - args.loss_tolerance:
            best_test_counter = 0
        best_test_loss = metrics_mean_dict['test_loss']
        best_test_iou = metrics_mean_dict['test_iou_coef']
        # best_test_class_list = []
        for idx_class, class_name in enumerate(class_names):
            best_test_iou_list[idx_class] = metrics_mean_dict[f'test_iou_coef_{class_name}']
        path_modelsave = f'./results/{args.sequence_name}/{args.run_name}'
        if not os.path.exists(f'{path_modelsave}/model'):
           # os.makedirs(f'{path_modelsave}')
           os.makedirs(f'{path_modelsave}/model')
        # torch.save(model.state_dict(), f'{path_modelsave}/model/model-{epoch:03d}.pt')
        # torch.save(model.state_dict(), f'{path_modelsave}/model/model-best.pt')
        model_module = model
        if is_data_parallel:
            model_module = model.module
        torch.save(model_module.state_dict(), f'{path_modelsave}/model/model-best.pt')

        # copy data.json to results folder
        shutil.copy2(f'{args.datasource_directory}/{args.json_filename}', path_modelsave)
    metrics_mean_dict['best_test_loss'] = best_test_loss
    metrics_mean_dict['best_test_iou'] = best_test_iou
    for idx_class, class_name in enumerate(class_names):
        metrics_mean_dict[f'best_test_iou_{class_name}'] = best_test_iou_list[idx_class]
    metrics_strs.append(f'best_test_loss: {round(best_test_loss, 2)}')
    metrics_strs.append(f'best_test_iou: {round(best_test_iou, 2)}')

    print(f'epoch: {epoch} {" ".join(metrics_strs)}, counter: {best_test_counter}')

    CsvUtils2.add_hparams(
        path_sequence=path_sequence,
        run_name=args.run_name,
        args_dict=args.__dict__,
        metrics_dict=metrics_mean_dict,  # contains train + test metrics
        global_step=epoch
    )

    if args.save_figs:
        make_figures(x, y, y_prim, class_names, epoch, metrics, args)

    if best_test_counter > args.early_stopping_patience:
        print(best_test_counter, best_test_loss, metrics['test_loss'][-args.early_stopping_patience:])
        break

