import torch.nn.functional as F
import torch.utils.data
import math


# from torchinfo import summary

class ResBlock(torch.nn.Module):
    def __init__(self, in_channels, out_channels, stride=1):
        super().__init__()
        self.conv1 = torch.nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride, padding=1, bias=False)
        self.gn1 = torch.nn.GroupNorm(num_channels=out_channels, num_groups=math.ceil(out_channels / 1))
        self.conv2 = torch.nn.Conv2d(out_channels, out_channels, kernel_size=3, stride=1, padding=1, bias=False)
        self.gn2 = torch.nn.GroupNorm(num_channels=out_channels, num_groups=math.ceil(out_channels / 1))
        self.is_bottleneck = False
        if stride != 1 or in_channels != out_channels:
            self.is_bottleneck = True
            self.shortcut = torch.nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=stride, bias=False)

    def forward(self, x):
        residual = x
        out = self.conv1.forward(x)
        out = F.relu(out)
        out = self.gn1.forward(out)
        out = self.conv2.forward(out)
        if self.is_bottleneck:
            residual = self.shortcut.forward(x)
        out += residual
        out = F.relu(out)
        out = self.gn2.forward(out)
        return out


class Model(torch.nn.Module):
    def __init__(self, classes=2, args=None):
        super().__init__()
        if args:
            coef_num_features = args.coef_num_features
            self.num_layers = args.num_layers
        else:
            coef_num_features = 3  # 1
            self.num_layers = 3

        self.maxpool = torch.nn.MaxPool2d(2)
        self.upsample = torch.nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)

        # self.conv_down_1 = ResBlock(in_channels=1, out_channels=int(8 * coef_num_features))
        # self.conv_down_2 = ResBlock(in_channels=int(8 * coef_num_features), out_channels=int(16 * coef_num_features))
        # self.conv_down_3 = ResBlock(in_channels=int(16 * coef_num_features), out_channels=int(32 * coef_num_features))
        # self.conv_down_4 = ResBlock(in_channels=int(32 * coef_num_features), out_channels=int(64 * coef_num_features))
        #
        # self.conv_middle = ResBlock(in_channels=int(64 * coef_num_features), out_channels=int(64 * coef_num_features))
        #
        # self.conv_up_4 = ResBlock(in_channels=int(128 * coef_num_features), out_channels=int(32 * coef_num_features))
        # self.conv_up_3 = ResBlock(in_channels=int(64 * coef_num_features), out_channels=int(16 * coef_num_features))
        # self.conv_up_2 = ResBlock(in_channels=int(32 * coef_num_features), out_channels=int(8 * coef_num_features))
        # self.conv_up_1 = ResBlock(in_channels=int(16 * coef_num_features), out_channels=classes)

        out_channels = int(2 ** coef_num_features)
        self.conv_down_1 = ResBlock(in_channels=1, out_channels=out_channels)
        self.conv_down_list: List[torch.nn.Module] = torch.nn.ModuleList()
        for idx_layer in range(self.num_layers):
            in_channels = int(2 ** (idx_layer + coef_num_features))
            out_channels = int(2 ** (idx_layer + coef_num_features + 1))
            self.conv_down_list.append(ResBlock(in_channels=in_channels, out_channels=out_channels))

        self.conv_middle = ResBlock(in_channels=out_channels, out_channels=out_channels)

        in_channels = int(2 ** (1 + coef_num_features))
        self.conv_up_1 = ResBlock(in_channels=in_channels, out_channels=classes)
        self.conv_up_list = torch.nn.ModuleList([])
        for idx_layer in range(self.num_layers):
            in_channels = int(2 ** (idx_layer + coef_num_features + 2))
            out_channels = int(2 ** (idx_layer + coef_num_features))
            self.conv_up_list.append(ResBlock(in_channels=in_channels, out_channels=out_channels))

    def forward(self, x):

        skip_1 = self.conv_down_1.forward(x)
        skip_list = []
        skip = skip_1
        for conv_down in self.conv_down_list:
            out = self.maxpool.forward(skip)
            skip = conv_down.forward(out)
            skip_list.append(skip)

        out = self.conv_middle.forward(skip)

        for idx in range(self.num_layers, 0, -1):
            skip = skip_list[idx - 1]
            conv_up = self.conv_up_list[idx - 1]
            out = torch.cat([out, skip], dim=1)
            out = conv_up.forward(out)
            out = self.upsample.forward(out)

        out = torch.cat([out, skip_1], dim=1)
        out = self.conv_up_1.forward(out)

        # skip_1 = self.conv_down_1.forward(x)
        # out = self.maxpool.forward(skip_1)
        #
        # skip_2 = self.conv_down_2.forward(out)
        # out = self.maxpool.forward(skip_2)
        #
        # skip_3 = self.conv_down_3.forward(out)
        # out = self.maxpool.forward(skip_3)
        #
        # skip_4 = self.conv_down_4.forward(out)
        #
        # out = self.conv_middle.forward(skip_4)
        #
        # out = torch.cat([out,  skip_4], dim=1)
        # out = self.conv_up_4.forward(out)
        # out = self.upsample.forward(out)
        #
        # out = torch.cat([out,  skip_3], dim=1)
        # out = self.conv_up_3.forward(out)
        # out = self.upsample.forward(out)
        #
        # out = torch.cat([out,  skip_2], dim=1)
        # out = self.conv_up_2.forward(out)
        # out = self.upsample.forward(out)
        #
        # out = torch.cat([out,  skip_1], dim=1)
        # out = self.conv_up_1.forward(out)

        # add correct activation function
        out = F.softmax(out, dim=1)
        return out

# if __name__ == '__main__':
#     import numpy as np
#     import argparse
#
#     parser = argparse.ArgumentParser(add_help=False)
#     # print(parser)
#     parser.add_argument('-num_layers', default=3, type=int)
#     parser.add_argument('-coef_num_features', default=1, type=int)
#     args, other_args = parser.parse_known_args()
#     # print(args)
#
#     classes = 5
#     model = Model(classes=classes, args=args)
#
#     params = list(model.parameters(recurse=True))
#     param_total = np.sum([np.prod(it.shape) for it in params])
#     print(f'param_total: {param_total}')
