# import torch
import torch.utils.data
import os
import json
import numpy as np
import torchvision.transforms as T


class CelebDataset(torch.utils.data.Dataset):
    def __init__(self, args):
        super(CelebDataset, self).__init__()

        # load mmap pointer
        mmap_dir = args.datasource_directory

        json_path = os.path.join(mmap_dir, 'data.json')
        with open(json_path) as fp:
            parameters = json.load(fp)

        # print(f'parameters: {parameters}')

        self.shape = tuple(parameters['shape'])
        self.channels = parameters['channels']

        mmap_path = os.path.join(mmap_dir, 'data.mmap')
        self.mmap = np.memmap(mmap_path, dtype='float16', mode='r', shape=self.shape)

        self.channel_grayscale = self.channels.index('photo_grayscale')
        self.channel_mask1 = self.channels.index('eye')
        # self.channel_mask1 = self.channels.index('lip')

        self.random_transformation = T.RandomApply(
            torch.nn.ModuleList([
                T.RandomHorizontalFlip(p=0.5),
                T.RandomCrop(size=(200, 200)),
                T.RandomAffine(degrees=(-70, 70), translate=(0.1, 0.3), scale=(0.8, 1.2)),
                T.Resize(size=self.shape[-1])
            ]), p=0.9)

    def __len__(self):
        return self.shape[0]  # get from json shape

    def __getitem__(self, idx):
        # x = torch.FloatTensor(np.array(self.mmap[idx, self.channel_grayscale:self.channel_grayscale + 1, :, :]))  # face
        # # y = torch.FloatTensor(np.array(self.mmap[idx, self.channel_mask1:self.channel_mask1+1, :, :])) # mask
        # y = torch.FloatTensor(np.array(self.mmap[idx, self.channel_mask1:, :, :]))  # mask
        # xy_cat = torch.cat((x, y), dim=0)  # concatenate face and mask

        xy = torch.FloatTensor(np.array(self.mmap[idx, self.channel_grayscale:, :, :]))
        xy_trans = self.random_transformation(xy)
        x = xy_trans[:1, :, :]

        # add random noise
        add_noise = AddRandomNoise(mean=(0., 0.), std=(0., 0.5))
        x = add_noise(x)

        # downsample and upsample (reduce quality)
        down_size = np.random.randint(int(self.shape[-1] / 8), int(self.shape[-1] / 2))
        resample = T.RandomApply(
            torch.nn.ModuleList([
                T.Resize(size=down_size),
                T.Resize(size=self.shape[-1])
            ]),
            p=0.9)
        x = resample(x)

        y = xy_trans[1:, :, :]
        y_other = torch.ones([1, y.shape[1], y.shape[2]]) - torch.sum(y, dim=0)
        y_other[y_other < 1] = 0
        y = torch.cat((y, y_other), dim=0)
        return x, y

    def getClassList(self):
        class_names = self.channels[self.channel_mask1:]
        class_names.append('other')
        return class_names

    def getInputChannels(self):
        return self.channels[self.channel_grayscale:self.channel_grayscale + 1]


class AddRandomNoise(object):
    def __init__(self, mean=(0., 1.), std=(0., 1.)):
        self.std = np.random.uniform(std[0], std[1])
        self.mean = np.random.uniform(mean[0], mean[1])

    def __call__(self, tensor):
        return tensor + torch.randn(tensor.size()) * self.std + self.mean

    def __repr__(self):
        return self.__class__.__name__ + '(mean={0}, std={1})'.format(self.mean, self.std)


if __name__ == '__main__':
    # import torch.utils.data
    import argparse

    parser = argparse.ArgumentParser(add_help=False)
    print(parser)

    parser.add_argument('-datasource_directory', default='../celeb_hq_dataset', type=str)
    parser.add_argument('-mmap_filename', default='data_mmap.dat', type=str)
    parser.add_argument('-json_filename', default='data_mmap.json', type=str)
    parser.add_argument('-learning_rate', default=1e-3, type=float)
    parser.add_argument('-batch_size', default=16, type=int)
    parser.add_argument('-num_layers', default=3, type=int)
    parser.add_argument('-sequence_name', default='seq_default', type=str)
    parser.add_argument('-run_name', default='run', type=str)
    parser.add_argument('-is_cuda', default=True, type=lambda x: (str(x).lower() == 'true'))
    parser.add_argument('-model', default='2_model_pre_trained_fcn', type=str)

    args, other_args = parser.parse_known_args()
    print(args)

    dataset_full = CelebDataset(args)
    print(dataset_full.shape)

    len_dataset = len(dataset_full)
    # print(dataset_full.shape)

    data_loader = torch.utils.data.DataLoader(
        dataset=dataset_full,
        batch_size=args.batch_size,
        shuffle=True,
        drop_last=False
    )
    # print(data_loader[0])

    for x, y in data_loader:
        print(x.size(), y.size())
